#include <time.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define talloc(type, num) (type *) malloc(sizeof(type)*(num))

//this is the useful struct which we will actually use in our program
//each tz_transition contains info about when the timezone changes, and the new timezone's offset from UTC (adjusted for DST)
//ts contains the GMT timestamp when the area's localtime changes
//offset contains the new offset from GMT for the area (in seconds)
typedef struct{
	long ts;
	long offset;
} tz_transition;

//tz_type structs represent how a timezone differs from GMT
//a tz_type has an offset from GMT, and dst may or may not apply
typedef struct {
	char tt_gmtoff[4];
	char tt_isdst[1];
	char tt_abbrind[1];
} tz_type_coded;

typedef struct {
	char    tzh_magic[4];       /* TZ_MAGIC */
	char    tzh_version[1];     /* '\0' or '2' as of 2005 */
	char    tzh_reserved[15];   /* reserved--must be zero */
	char    tzh_ttisgmtcnt[4];  /* coded number of trans. time flags */
	char    tzh_ttisstdcnt[4];  /* coded number of trans. time flags */
	char    tzh_leapcnt[4];     /* coded number of leap seconds */
	char    tzh_timecnt[4];     /* coded number of transition times */
	char    tzh_typecnt[4];     /* coded number of local time types */
	char    tzh_charcnt[4];     /* coded number of abbr. chars */
} tzhead;

//The following functions convert a nbyte long char * to a type (long,int,etc...)
long convert_bytes_long(const char * codep,int nbytes){
	long result;
	int i;

	result = (codep[0] & 0x80) ? ~0L : 0;
	for (i = 0; i < nbytes; ++i)
		result = (result << 8) | (codep[i] & 0xff);
	return result;
}

int convert_bytes_int(const char * codep,int nbytes){
	int result;
	int i;

	result = (codep[0] & 0x80) ? ~0L : 0;
	for (i = 0; i < nbytes; ++i)
		result = (result << 8) | (codep[i] & 0xff);
	return result;
}

int convert_bytes_uint(const char * codep,int nbytes){
	unsigned int result;
	int i;

	result = (codep[0] & 0x80) ? ~0L : 0;
	for (i = 0; i < nbytes; ++i)
		result = (result << 8) | (codep[i] & 0xff);
	return result;
}

//frees array containing tz_transitions
void free_trans_info(tz_transition ** arr_tz_trans){

	int i;
	i=0;

	while(arr_tz_trans[i] != NULL){
		free(arr_tz_trans[i]);	
		i++;
	}
	free(arr_tz_trans);
}

//first entry in our array of tz_transitions contains the first known offset after time ts_start
tz_transition ** get_tz_transitions(char * cstr_tz, time_t ts_start){

	int i;

	char *cstr_file;
	char *file_loc;
	FILE * file_tz;

	int index_type,index_trans;

	//used with fread
	tz_type_coded tz_st_coded; //timezone type (coded in bytes)
	tzhead header;
	char bytes4[4];
	char c;

	long tzh_timecnt;//# of transition times
	long tzh_typecnt;//# of different timezone types/types

	long trans_time;
	long first_offset = -1;
	int found_first_offset = 0;	
	long offset;
	
	long * arr_offsets;
	int * arr_indices;

	tz_transition * tz_trans;
	tz_transition * tz_trans_first;
	tz_transition ** arr_tz_trans;

	//file stream position indicators
	long pos_types;
	long pos_indices;
	long pos_trans;
	
	file_loc = (char *)"/usr/share/zoneinfo/";

	//no timezone name should be longer than 300 chars (just a guess, but probably more than enough)
	cstr_file = talloc(char,strlen(file_loc)+300);

	strcpy(cstr_file,file_loc);
	strncat(cstr_file,cstr_tz,299);

	file_tz = fopen(cstr_file,"r");
	if(!file_tz){
		//the file could not be opened
		free(cstr_file);
		return NULL; 
	}

	//read the header
	fread(&header,sizeof(tzhead),1,file_tz);

	//get the number of transition times and number of offsets (types)
	tzh_typecnt = convert_bytes_long(header.tzh_typecnt,4);
	tzh_timecnt = convert_bytes_long(header.tzh_timecnt,4);
	
	//make sure this is a version 2 file
	if(header.tzh_version[0] != '2'){
		//the timezone file is not a version 2 file
		free(cstr_file);
		fclose(file_tz);
		return NULL; 
	}

	//store the position right before the transition times
	pos_trans = ftell(file_tz);

	//store the position before the indices
	pos_indices = pos_trans + tzh_timecnt * 4; 

	//store position before the types
	pos_types = pos_indices + tzh_timecnt;

	//read all indices
	fseek(file_tz, pos_indices,SEEK_SET);
	arr_indices = talloc(int,tzh_timecnt);
	for(i=0;i<tzh_timecnt;i++){
		//read 1 byte
		fread(&c,1,1,file_tz);
		arr_indices[i] = (int)c;
	}

	//now read all types
	fseek(file_tz, pos_types,SEEK_SET);
	arr_offsets = talloc(long,tzh_typecnt);
	for(i=0;i<tzh_typecnt;i++){
		fread(&tz_st_coded,sizeof(tz_type_coded),1,file_tz);
		//convert and store the offset
		arr_offsets[i] = convert_bytes_long(tz_st_coded.tt_gmtoff,4);
		//check if this is the first non-dst type
		if(first_offset == -1 && !(int)tz_st_coded.tt_isdst[0]){
			first_offset = arr_offsets[i];
		}
	}
	//there was no non-dst type, so the first known offset is the first listed in the file
	if(first_offset == -1){
		first_offset = arr_offsets[0];
	}

	//use the indices and types to match transition times to offsets
	fseek(file_tz, pos_trans,SEEK_SET);

	//the first element is the earliest known timezone > our input time
	//then, there are tzh_timecnt transitions
	//finally, a NULL terminator
	arr_tz_trans = talloc(tz_transition *,tzh_timecnt+2);

	//create the first element
	tz_trans_first = talloc(tz_transition, 1);
	tz_trans_first->ts = -1; //we don't know when the time was set to this offset, but it is the first
	tz_trans_first->offset = first_offset;
	arr_tz_trans[0] = tz_trans_first;

	index_trans = 1;//where we are inserting into arr_tz_trans
	for(i=0;i<tzh_timecnt;i++){
		//read 4 bytes
		fread(bytes4,4,1,file_tz);

		index_type = arr_indices[i];
		offset = arr_offsets[index_type];
		trans_time = convert_bytes_long(bytes4,4);

		if(trans_time > ts_start){
			if(!found_first_offset){
				found_first_offset = 1;	
			}
			//make sure that this offset does not equal the previous offset
			//sometimes the timezone files have "transitions"
			//where the GMT offset is the same, and only the timezone abbreviation changes
			if(tz_trans != NULL && tz_trans->offset != offset){
				//store this transition
				tz_trans = talloc(tz_transition, 1);
				tz_trans->ts = trans_time;
				tz_trans->offset = offset;
				arr_tz_trans[index_trans] = tz_trans;
				index_trans++;
			}
		}else{
			tz_trans_first->offset = offset;
			tz_trans_first->ts = trans_time;
		}
	}
	
	//this is a signal for the end of our array
	arr_tz_trans[index_trans] = NULL;
	
	//cleanup
	fclose(file_tz);
	free(arr_indices); 
	free(arr_offsets);
	free(cstr_file);

	return arr_tz_trans;

}

int main(int argc, char ** argv){

	int i;
	tz_transition ** tz_transitions;
	tz_transition * tz_trans, *tz_trans_prev;
	char *cstr_ts;
	char *cstr_ts_global;
	struct tm * timeinfo;
	struct tm * timeinfo_global;
	time_t ts;
	time_t ts_global;
	time_t ts_cur;

	if(argc != 3){
		printf("Usage: tz_transitions 'timezone' 'ts-start'\n");
		printf("       or\n");
		printf("       tz_transitions 'timezone' ""now""\n");
		return 0;
	}

	if(strcmp(argv[2],"now")){
		ts_cur = atoi(argv[2]);
	}else{
		ts_cur = time(NULL);
	}

	tz_transitions = get_tz_transitions(argv[1],ts_cur);

	if(!tz_transitions){
		//could not get the transitions
		printf("Error: could not get the transitions for timezone: %s\n",argv[1]);
		return 0;
	}
	
	//print the transitions
	i=0;
	cstr_ts = talloc(char, 30);
	cstr_ts_global = talloc(char, 30);
	
	//print out the data
	printf("The starting timezone is GMT");
	tz_trans_prev = tz_transitions[i];
	if(tz_trans_prev->offset > 0){
		printf("+");
	}
	printf("%G\n\n",(float)tz_trans_prev->offset/3600);
	i++;
	while(tz_transitions[i] != NULL){
		tz_trans = tz_transitions[i];
		tz_trans_prev = tz_transitions[i-1];
		ts = tz_trans->ts; //timestamp for the transition
		ts_global = tz_trans->ts;

		//make ts GMT by subtracting the last offset
		//ts is the GMT time when the transition occurs
		//convert the GMT time to the current localtime using the last known time offset
		//this is done so we can use gmtime and strftime
		//those 2 functions interpret the time as GMT, so we want GMT to really be the localtime
		//it is confusing...
		ts += tz_trans_prev->offset;

		timeinfo = gmtime(&ts);

		strftime(cstr_ts,30,"%B %d, %Y %H:%M:%S",timeinfo);
		
		timeinfo_global = gmtime(&ts_global);

		//convert transition time to an easily read format
		strftime(cstr_ts_global,30,"%B %d, %Y %H:%M:%S",timeinfo_global);

		printf("Transition time in file:%ld\nAdd previous offset to get local transition time: %s\nNew timezone is GMT",ts_global,cstr_ts);
		if(tz_trans->offset > 0){
			printf("+");
		}
		printf("%G\n",(float)tz_trans->offset/3600);
		i++;
		printf("\n");
	}
	
	//free memory
	free_trans_info(tz_transitions);
	free(cstr_ts);

	return 0;
}
